'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var ts = require('typescript');
var ts__default = _interopDefault(ts);
var child_process = require('child_process');
var path = _interopDefault(require('path'));

/**
 * Helper function to create an arrow function with one argument
 *
 * @example
 * ```
 * // If we create the function like this
 * const arrowFunction = createSingleParameterFunction(
 *   (argument) => ts.createStrictEquality(argument, ts.createNull())
 * );
 *
 * // The generated code will look like this
 * (value) => value === null;
 * ```
 *
 * @example
 * ```
 * // Sometimes the argument name would shadow a variable from parent scope
 * // In those case we can rename it to something else than 'value'
 * const arrowFunction = createSingleParameterFunction(
 *   (argument) => ts.createStrictEquality(argument, ts.createNull()),
 *   'element'
 * );
 *
 * // The generated code will look like this
 * (element) => element === null;
 * ```
 *
 * @param body {FunctionBodyCreator<[ts.Identifier]>} Function that accepts the argument identifier and returns the function body
 * @param argumentName {String} [value] Name of the function argument
 */
const createSingleParameterFunction = (body, argumentName = 'value') => {
    const argument = ts__default.createIdentifier(argumentName);
    return ts__default.createArrowFunction(
    /* modifiers */ undefined, 
    /* typeParameters */ undefined, [
        ts__default.createParameter(
        /* decorators */ undefined, 
        /* modifiers */ undefined, 
        /* dotDotDotToken */ undefined, 
        /* name */ argument),
    ], undefined, ts__default.createToken(ts__default.SyntaxKind.EqualsGreaterThanToken), body(argument));
};
/**
 * Helper function for accessing object properties
 *
 * @example
 * ```
 * const identifier = ts.createIdentifier('obj');
 *
 * const accessKey = createElementAccess(identifier, 'property');
 * const accessElement = createElementAccess(identifier, 1);
 *
 * // The generated code will look like this
 * obj['property']
 * obj[1]
 * ```
 *
 * @param value {ts.Expression} The object which property should be accessed
 * @param property {String | Number} The property name / element index
 */
const createElementAccess = (value, property) => ts__default.createElementAccess(value, ts__default.createLiteral(property.toString()));
const createObjectWithProperties = (properties) => ts__default.createObjectLiteral(properties, true);
const createVariable = (identifier, initializer) => ts__default.createVariableStatement(undefined, [ts__default.createVariableDeclaration(identifier, undefined, initializer)]);
const createArrayEvery = (value, body, callbackArgumentName = 'element') => ts__default.createCall(ts__default.createPropertyAccess(value, 'every'), [], [createSingleParameterFunction(body, callbackArgumentName)]);
const createArrayFrom = (value) => ts__default.createCall(ts__default.createPropertyAccess(ts__default.createIdentifier('Array'), 'from'), [], [value]);
const createIsArray = (value) => ts__default.createCall(ts__default.createPropertyAccess(ts__default.createIdentifier('Array'), 'isArray'), [], [value]);
const createObjectKeys = (value) => ts__default.createCall(ts__default.createPropertyAccess(ts__default.createIdentifier('Object'), 'keys'), [], [value]);
const parenthesize = (expressions) => expressions.map((expression) => ts__default.isCallExpression(expression) ? expression : ts__default.createParen(expression));
const createLogicalAndChain = (...expressions) => {
    return parenthesize(expressions).reduce((chain, expression) => ts__default.createLogicalAnd(chain, expression));
};
const createLogicalOrChain = (...expressions) => {
    return parenthesize(expressions).reduce((chain, expression) => ts__default.createLogicalOr(chain, expression));
};
const createIsOfType = (value, type) => ts__default.createStrictEquality(ts__default.createTypeOf(value), type);
const createIsInstanceOf = (value, className) => ts__default.createBinary(value, ts__default.SyntaxKind.InstanceOfKeyword, className);
const createDoubleNegation = (value) => ts__default.createPrefix(ts__default.SyntaxKind.ExclamationToken, ts__default.createPrefix(ts__default.SyntaxKind.ExclamationToken, value));
// See
//
// https://www.typescriptlang.org/docs/handbook/release-notes/typescript-2-2.html#object-type
//
// for more information
const createIsNotPrimitive = (value) => {
    return createLogicalOrChain(createIsOfType(value, ts__default.createLiteral('function')), createLogicalAndChain(createIsOfType(value, ts__default.createLiteral('object')), createDoubleNegation(value)));
};
const createIsNotNullOrUndefined = (value) => createLogicalAndChain(ts__default.createStrictInequality(value, ts__default.createIdentifier('undefined')), ts__default.createStrictInequality(value, ts__default.createNull()));
const createRequire = (identifier, path, property = 'default') => createVariable(identifier, ts__default.createPropertyAccess(ts__default.createCall(ts__default.createIdentifier('require'), undefined, [ts__default.createLiteral(path)]), property));

/**
 * Hash (as in plain JavaScript object) based implementation of TypeGuardRegistry
 */
class HashTypeGuardRegistry {
    constructor(identifier) {
        this.identifier = identifier;
        this.typeGuardsBeingCreated = new Set();
        this.typeGuardFunctionsByTypeName = new Map();
        this.cyclicTypeNames = new Set();
    }
    get(typeName) {
        // First we check that a type under this name has been registered
        if (!this.typeGuardFunctionsByTypeName.has(typeName))
            return undefined;
        // Then we create a runtime reference to the type guard
        return createElementAccess(this.identifier, typeName);
    }
    create(typeName, factory) {
        // Get an existing type guard function expression
        const hasTypeGuard = this.typeGuardFunctionsByTypeName.has(typeName);
        // Check whether we are not in progress of creating the very same type guard
        //
        // This can happen when types reference themselves circularly, for example:
        //
        // interface MyInterface {
        //   parent: MyInterface | undefined;
        // }
        //
        const typeGuardIsBeingCreated = this.typeGuardsBeingCreated.has(typeName);
        // If the type guard with the same name is being created it means
        // that the type is cyclic so we mark it as such
        if (typeGuardIsBeingCreated) {
            this.cyclicTypeNames.add(typeName);
        }
        // If we don't have the type guard yet and we are not creating one at the moment
        // we need to create it
        if (!hasTypeGuard && !typeGuardIsBeingCreated) {
            // First let's mark the fact that we are creating it now
            this.typeGuardsBeingCreated.add(typeName);
            // Then create the type guard and store it in the map
            this.typeGuardFunctionsByTypeName.set(typeName, createSingleParameterFunction(factory));
            // Finally let's remove the mark
            this.typeGuardsBeingCreated.delete(typeName);
        }
        // We return an expression that points to the type guard function
        return createElementAccess(this.identifier, typeName);
    }
    code() {
        // There is no need for a object if it is not used in the code
        if (this.typeGuardFunctionsByTypeName.size === 0)
            return [];
        // We will use this to remember whether there were any cyclic types
        let areAnyTypesCyclic = false;
        // In order not to create the cycle breaker in every file we import it from our module
        const wrapperIdentifier = ts__default.createIdentifier('__typeGuardCycleBreaker__');
        const wrapperImport = createRequire(wrapperIdentifier, 'ts-type-checked/runtime', 'typeGuardCycleBreaker');
        const wrapTypeGuard = (typeGuard) => ts__default.createCall(wrapperIdentifier, undefined, [typeGuard]);
        // Now we create an object literal with all the type guards keyed by type names
        const typeGuardEntries = Array.from(this.typeGuardFunctionsByTypeName.entries());
        const properties = typeGuardEntries.map(([typeName, typeGuard]) => {
            // First we check whether the type has been marked as cyclic
            const isCyclic = this.cyclicTypeNames.has(typeName);
            // Make sure we remember meeting any cyclic types
            areAnyTypesCyclic = areAnyTypesCyclic || isCyclic;
            // If the type is not cyclic we output a simple property assignment,
            // if it is cyclic we need to wrap the type guard in a cycle-breaking code
            const wrappedTypeGuard = isCyclic ? wrapTypeGuard(typeGuard) : typeGuard;
            return ts__default.createPropertyAssignment(ts__default.createLiteral(typeName), wrappedTypeGuard);
        });
        return [
            // If there were any cyclic types we will need to import the cycle breaker
            ...(areAnyTypesCyclic ? [wrapperImport] : []),
            createVariable(this.identifier, createObjectWithProperties(properties)),
        ];
    }
}

/**
 * Map based implementation of TypeDescriptorRegistry
 */
class MapTypeDescriptorRegistry {
    // We will also need a TypeNameGenerator to create names for types
    constructor(typeNameGenerator) {
        this.typeNameGenerator = typeNameGenerator;
        // Type to TypeName map
        this.typeNamesByType = new Map();
        // TypeName to TypeDescriptor map
        this.typeDescriptorsByTypeName = new Map();
    }
    get(typeName) {
        // Simply get the type descriptor from the map
        return this.typeDescriptorsByTypeName.get(typeName);
    }
    create(type, factory) {
        // First check whether we already have a type name for this type.
        const existingTypeName = this.typeNamesByType.get(type);
        if (existingTypeName !== undefined)
            return existingTypeName;
        // Generate a name for this type
        const typeName = this.typeNameGenerator(type);
        // First we store the name in the map which effectively "marks" the type as resolved
        this.typeNamesByType.set(type, typeName);
        // Only then we create the type descriptor. This is important since if we didn't
        // store the typeName in typeNamesByType prior to creating the descriptor
        // we might end up spinning in type reference circles
        this.typeDescriptorsByTypeName.set(typeName, factory());
        return typeName;
    }
}

const isDevelopment = process.env.NODE_ENVIRONMENT === 'development';
const defaultTransformerOptions = {
    mode: isDevelopment ? 'development' : 'production',
    logLevel: 'normal',
};

const createDebugTypeNameGenerator = (typeChecker) => {
    const existingNames = new Set();
    return (type) => {
        const originalTypeName = typeChecker.typeToString(type);
        let typeName = originalTypeName;
        let attempt = 1;
        while (existingNames.has(typeName)) {
            typeName = originalTypeName + '~' + ++attempt;
        }
        existingNames.add(typeName);
        return typeName;
    };
};

var LogFeature;
(function (LogFeature) {
    LogFeature[LogFeature["DEBUG"] = 1] = "DEBUG";
    LogFeature[LogFeature["INFO"] = 2] = "INFO";
    LogFeature[LogFeature["WARN"] = 4] = "WARN";
    LogFeature[LogFeature["ERROR"] = 8] = "ERROR";
})(LogFeature || (LogFeature = {}));
const LOG_FEATURES = {
    debug: LogFeature.DEBUG | LogFeature.INFO | LogFeature.WARN | LogFeature.ERROR,
    nosey: LogFeature.INFO | LogFeature.WARN | LogFeature.ERROR,
    normal: LogFeature.WARN | LogFeature.ERROR,
    silent: 0,
};
const noop = () => undefined;
const bindConsole = (method, prefix) => console[method].bind(console, ...prefix);
const createLogger = (logLevel, ...prefix) => {
    const logFeatures = LOG_FEATURES[logLevel];
    return {
        debug: logFeatures & LogFeature.DEBUG ? bindConsole('debug', prefix) : noop,
        info: logFeatures & LogFeature.INFO ? bindConsole('info', prefix) : noop,
        warn: logFeatures & LogFeature.WARN ? bindConsole('warn', prefix) : noop,
        error: logFeatures & LogFeature.ERROR ? bindConsole('error', prefix) : noop,
        indent: () => createLogger(logLevel, ...prefix, '\t'),
    };
};

const createProductionTypeNameGenerator = () => {
    const existingNames = new Set();
    return () => {
        let typeName = existingNames.size;
        while (existingNames.has(typeName))
            typeName++;
        existingNames.add(typeName);
        return String(typeName);
    };
};

const INDEX_JS = path.join(__dirname, 'index.js');
const INDEX_TS = path.join(__dirname, 'index.d.ts');
const PWD = child_process.execSync('pwd').toString();
const isOurImportExpression = (node) => {
    if (!ts__default.isImportDeclaration(node))
        return false;
    const sourceFile = node.getSourceFile().fileName;
    const module = node.moduleSpecifier.text;
    const isModulePathRelative = module.startsWith('.');
    const modulePath = isModulePathRelative ? path.resolve(path.dirname(sourceFile), module) : module;
    try {
        const resolvedPath = require.resolve(modulePath, { paths: [PWD, sourceFile] });
        return INDEX_JS === resolvedPath;
    }
    catch (e) {
        return false;
    }
};
const isJSDocSignature = (declaration) => {
    if (typeof ts__default.isJSDocSignature !== 'function')
        return false;
    if (!declaration)
        return false;
    return ts__default.isJSDocSignature(declaration);
};
const isOurCallExpression = (node, name, typeChecker) => {
    var _a, _b;
    if (!ts__default.isCallExpression(node))
        return false;
    const declaration = (_a = typeChecker.getResolvedSignature(node)) === null || _a === void 0 ? void 0 : _a.declaration;
    return (!!declaration &&
        // Declaration must be there
        !isJSDocSignature(declaration) &&
        // It has to come from our .d.ts definition file
        path.join(declaration.getSourceFile().fileName) === INDEX_TS &&
        // And its name must match the expected name
        ((_b = declaration.name) === null || _b === void 0 ? void 0 : _b.getText()) === name);
};

/**
 * Factory for ASTVisitor that replaces occurrences of isA and typeCheckFor
 * with generated type guards
 *
 * This visitor inspects the code and uses the typeGuardResolver to generate the type guards
 *
 * @param typeChecker {ts.TypeChecker} Instance of TypeChecker
 * @param typeGuardResolver {TypeCheckFactory} Function that turns a type into a type guard expression
 */
const createTypeCheckVisitor = (typeChecker, typeGuardResolver) => {
    return (node) => {
        var _a, _b;
        // All the imports from this module are fake so we need to remove them all
        if (isOurImportExpression(node))
            return undefined;
        if (isOurCallExpression(node, 'isA', typeChecker)) {
            const typeNode = (_a = node.typeArguments) === null || _a === void 0 ? void 0 : _a[0];
            if (!typeNode) {
                throw new Error('isA<T>() requires one type parameter, none specified');
            }
            const valueNode = node.arguments[0];
            if (!valueNode) {
                throw new Error('isA<T>() requires one argument, none specified');
            }
            return typeGuardResolver(typeNode, valueNode);
        }
        if (isOurCallExpression(node, 'typeCheckFor', typeChecker)) {
            const typeNode = (_b = node.typeArguments) === null || _b === void 0 ? void 0 : _b[0];
            if (!typeNode) {
                throw new Error('typeCheckFor<T>() requires one type parameter, none specified');
            }
            return createSingleParameterFunction((value) => typeGuardResolver(typeNode, value));
        }
        return node;
    };
};

const isBigInt = (type, libraryDescriptorName) => !!(type.flags & ts__default.TypeFlags.BigInt) || libraryDescriptorName === 'BigInt';
const isBoolean = (type, libraryDescriptorName) => !!(type.flags & ts__default.TypeFlags.Boolean) || libraryDescriptorName === 'Boolean';
const isString = (type, libraryDescriptorName) => !!(type.flags & ts__default.TypeFlags.String) || libraryDescriptorName === 'String';
const isNumber = (type, libraryDescriptorName) => !!(type.flags & ts__default.TypeFlags.Number) || libraryDescriptorName === 'Number';
const isSymbol = (type, libraryDescriptorName) => !!(type.flags & ts__default.TypeFlags.ESSymbol) || libraryDescriptorName === 'Symbol';
const isDate = (type, libraryDescriptorName) => libraryDescriptorName === 'Date';
const isMap = (type, libraryDescriptorName) => libraryDescriptorName === 'Map';
const isSet = (type, libraryDescriptorName) => libraryDescriptorName === 'Set';
const isPromise = (type, libraryDescriptorName) => libraryDescriptorName === 'Promise';
const isRegExp = (type, libraryDescriptorName) => libraryDescriptorName === 'RegExp';
const isInterface = (type, libraryDescriptorName) => !!(type.flags & ts__default.TypeFlags.Object) || libraryDescriptorName === 'Object';
const isLiteral = (type) => (typeof type.isLiteral === 'function' && type.isLiteral()) ||
    !!(type.flags & ts__default.TypeFlags.BigIntLiteral) ||
    !!(type.flags & ts__default.TypeFlags.Literal);
const isNull = (type) => !!(type.flags & ts__default.TypeFlags.Null);
const isUndefined = (type) => !!(type.flags & ts__default.TypeFlags.Undefined || type.flags & ts__default.TypeFlags.Void);
const isAny = (type) => !!(type.flags & ts__default.TypeFlags.Any || type.flags & ts__default.TypeFlags.Unknown);
const isNever = (type) => !!(type.flags & ts__default.TypeFlags.Never);
const isObjectKeyword = (typeNode) => (typeNode === null || typeNode === void 0 ? void 0 : typeNode.kind) === ts__default.SyntaxKind.ObjectKeyword;
const isTrueKeyword = (typeNode) => (typeNode === null || typeNode === void 0 ? void 0 : typeNode.kind) === ts__default.SyntaxKind.TrueKeyword;
const isFalseKeyword = (typeNode) => (typeNode === null || typeNode === void 0 ? void 0 : typeNode.kind) === ts__default.SyntaxKind.FalseKeyword;
const isTuple = (type, typeNode) => (typeNode === null || typeNode === void 0 ? void 0 : typeNode.kind) === ts__default.SyntaxKind.TupleType;
const isIntersection = (type, typeNode) => (typeof type.isIntersection === 'function' && type.isIntersection()) ||
    (typeNode === null || typeNode === void 0 ? void 0 : typeNode.kind) === ts__default.SyntaxKind.IntersectionType;
const isUnion = (type, typeNode) => (typeof type.isUnion === 'function' && type.isUnion()) ||
    (typeNode === null || typeNode === void 0 ? void 0 : typeNode.kind) === ts__default.SyntaxKind.UnionType ||
    !!(type.getFlags() & ts__default.TypeFlags.Union);
const isFunction = (type, libraryDescriptorName, typeNode) => {
    var _a, _b;
    return (typeNode === null || typeNode === void 0 ? void 0 : typeNode.kind) === ts__default.SyntaxKind.FunctionType ||
        (typeNode === null || typeNode === void 0 ? void 0 : typeNode.kind) === ts__default.SyntaxKind.ConstructorType ||
        libraryDescriptorName === 'Function' ||
        !!((_a = type.getConstructSignatures()) === null || _a === void 0 ? void 0 : _a.length) ||
        !!((_b = type.getCallSignatures()) === null || _b === void 0 ? void 0 : _b.length);
};
const isArrayType = (typeChecker, type) => { var _a; return typeof typeChecker.isArrayType === 'function' && !!((_a = typeChecker) === null || _a === void 0 ? void 0 : _a.isArrayType(type)); };
const isArray = (typeChecker, type, libraryDescriptorName, typeNode) => (typeNode === null || typeNode === void 0 ? void 0 : typeNode.kind) === ts__default.SyntaxKind.ArrayType || libraryDescriptorName === 'Array' || isArrayType(typeChecker, type);

const functionTypeWarning = (typeName) => `

It looks like you are trying to type check a function-like value (${typeName}). 
Due to very nature of JavaScript it's not possible to see what the return type of a function is
or what the signature of a function was.

ts-type-checked can only check whether something is of type function, nothing more. Sorry :(

`;
const promiseTypeWarning = (typeName) => `

It looks like you are trying to type check a Promise-like value (${typeName}). 
Although possible, type checking Promises is discouraged in favour of wrapping the value in a new Promise:

const certainlyPromise = Promise.resolve(value);

Check https://stackoverflow.com/questions/27746304/how-do-i-tell-if-an-object-is-a-promise for more information.

`;

const getFirstValidDeclaration = (declarations) => {
    return ((declarations === null || declarations === void 0 ? void 0 : declarations.find((declaration) => !ts__default.isVariableDeclaration(declaration) && !ts__default.isFunctionDeclaration(declaration))) || (declarations === null || declarations === void 0 ? void 0 : declarations[0]));
};

const typeDescriptorNameBySymbolName = {
    Array: 'Array',
    ReadonlyArray: 'Array',
    BigInt: 'BigInt',
    Number: 'Number',
    Function: 'Function',
    Date: 'Date',
    String: 'String',
    Boolean: 'Boolean',
    Object: 'Object',
    Promise: 'Promise',
    RegExp: 'RegExp',
    Map: 'Map',
    Set: 'Set',
    Symbol: 'Symbol',
};
const isSourceFileDefaultLibrary = (program, file) => {
    if (program.isSourceFileDefaultLibrary(file))
        return true;
    if (file.fileName.match(/typescript\/lib\/lib\..*\.d\.ts$/))
        return true;
    return false;
};
const getLibraryTypeDescriptorName = (program, type) => {
    var _a, _b;
    const declaration = getFirstValidDeclaration((_a = type.symbol) === null || _a === void 0 ? void 0 : _a.declarations);
    const sourceFile = declaration === null || declaration === void 0 ? void 0 : declaration.getSourceFile();
    if (!sourceFile || !isSourceFileDefaultLibrary(program, sourceFile))
        return undefined;
    return typeDescriptorNameBySymbolName[(_b = type.symbol) === null || _b === void 0 ? void 0 : _b.name];
};

const getDOMElementClassName = (program, type) => {
    var _a, _b;
    const declaration = getFirstValidDeclaration((_a = type.symbol) === null || _a === void 0 ? void 0 : _a.declarations);
    const sourceFile = declaration === null || declaration === void 0 ? void 0 : declaration.getSourceFile();
    if (!sourceFile || !isSourceFileDefaultLibrary(program, sourceFile))
        return undefined;
    if (!sourceFile.fileName.match(/lib.dom.d.ts$/))
        return undefined;
    if (!((_b = type.symbol) === null || _b === void 0 ? void 0 : _b.name.match(/(Element|^Document|^Node)$/i)))
        return undefined;
    return type.symbol.name;
};

/**
 * Helper function that checks whether the array of modifiers
 * contains "private" or "protected" keywords.
 *
 * @param modifiers {ts.ModifiersArray} [undefined] The array of modifiers
 */
const hasPrivateOrProtectedModifiers = (modifiers) => !!(modifiers === null || modifiers === void 0 ? void 0 : modifiers.some((modifier) => modifier.kind === ts__default.SyntaxKind.PrivateKeyword || modifier.kind === ts__default.SyntaxKind.ProtectedKeyword));
const isPrivateIdentifier = (node) => typeof ts__default.isPrivateIdentifier === 'function' ? ts__default.isPrivateIdentifier(node) : false;
/**
 * Helper function that checks whether a property represented by a Symbol
 * is publicly visible, i.e. it does not have "private" or "protected" modifier
 *
 * @param property {ts.Symbol} Property symbol
 */
const isPublicProperty = (property) => {
    const declaration = property.valueDeclaration;
    if (!declaration) {
        // TODO This is just a "guess", maybe the missing declaration can mean a private/protected property
        return true;
    }
    if (ts__default.isPropertySignature(declaration) ||
        ts__default.isPropertyDeclaration(declaration) ||
        ts__default.isMethodDeclaration(declaration) ||
        ts__default.isMethodSignature(declaration) ||
        ts__default.isParameter(declaration) ||
        ts__default.isGetAccessor(declaration)) {
        if (isPrivateIdentifier(declaration.name))
            return false;
        return !hasPrivateOrProtectedModifiers(declaration.modifiers);
    }
    return false;
};
/**
 * Helper function that return property name as a ts.Expression.
 * It will make sure that is the property is a numeric literal,
 * it is returned as a number rather than a number-like string
 *
 * @param property {ts.Symbol} The property to get the name of
 * @param typeChecker {ts.TypeChecker} Instance of ts.TypeChecker
 * @param scope {ts.TypeNode} The root TypeNode that contained the type
 */
const getPropertyName = (property, typeChecker, scope) => {
    // Let's get the property type
    const propertyType = 
    // The nameType property is not documented but can serve as a good starting point,
    // saves one function call :)
    property.nameType || typeChecker.getTypeOfSymbolAtLocation(property, scope);
    // If the property type exists and it looks like a number literal then let's turn it into a number
    if (propertyType && typeof propertyType.flags === 'number' && propertyType.flags & ts__default.TypeFlags.NumberLiteral) {
        const nameAsNumber = parseFloat(property.name);
        if (!isNaN(nameAsNumber) && String(nameAsNumber) === property.name) {
            return ts__default.createLiteral(nameAsNumber);
        }
    }
    return ts__default.createLiteral(property.name);
};
/**
 * Helper function that returns a property accessor - either a property name (e.g. 'name')
 * or a computed property expression (e.g. Symbol.toStringTag)
 *
 * @param property {ts.Symbol} The property to get the accessor of
 * @param typeChecker {ts.TypeChecker} Instance of ts.TypeChecker
 * @param scope {ts.TypeNode} The root TypeNode that contained the type
 */
const getPropertyAccessor = (property, typeChecker, scope) => {
    const declaration = property.valueDeclaration;
    if (declaration &&
        (ts__default.isPropertySignature(declaration) ||
            ts__default.isPropertyDeclaration(declaration) ||
            ts__default.isMethodDeclaration(declaration) ||
            ts__default.isMethodSignature(declaration))) {
        if (ts__default.isComputedPropertyName(declaration.name)) {
            return ts__default.createIdentifier(declaration.name.expression.getFullText());
        }
    }
    return getPropertyName(property, typeChecker, scope);
};

const getPropertyTypeDescriptors = (typeChecker, scope, properties, typeNameResolver) => {
    return properties.filter(isPublicProperty).map((property) => {
        const propertyType = typeChecker.getTypeOfSymbolAtLocation(property, scope);
        const accessor = getPropertyAccessor(property, typeChecker, scope);
        return {
            _type: 'property',
            accessor,
            type: typeNameResolver(scope, propertyType),
        };
    });
};

/**
 * A factory for TypeDescriptorGenerator functions.
 *
 * TypeDescriptorGenerator functions accept a TypeNode and a Type and spit out
 * a TypeDescriptor (if the type is simple) or a TypeDescriptorGeneratorCallback if the type
 * contains nested types that also need to be resolved into a TypeDescriptor.
 *
 * We want to be able to intercept the calls to TypeDescriptorGenerator in order to break dependency cycles
 * so it can't just call itself when it needs to resolve a nested type, instead it returns
 * a TypeDescriptorGeneratorCallback which gives this function access to
 *
 * @param program {ts.Program} A TypeScript Program instance
 * @param logger {Logger} An instance of Logger to use
 */
const createTypeDescriptorGenerator = (program, logger) => (scope, type) => {
    var _a;
    const typeChecker = program.getTypeChecker();
    const libraryDescriptorName = getLibraryTypeDescriptorName(program, type);
    // BigInt
    if (isBigInt(type, libraryDescriptorName)) {
        logger.debug('BigInt');
        return { _type: 'keyword', value: 'bigint' };
    }
    // Boolean
    if (isBoolean(type, libraryDescriptorName)) {
        logger.debug('Boolean');
        return { _type: 'keyword', value: 'boolean' };
    }
    // Number
    if (isNumber(type, libraryDescriptorName)) {
        logger.debug('Number');
        return { _type: 'keyword', value: 'number' };
    }
    // String
    if (isString(type, libraryDescriptorName)) {
        logger.debug('String');
        return { _type: 'keyword', value: 'string' };
    }
    // Symbol
    if (isSymbol(type, libraryDescriptorName)) {
        logger.debug('Symbol');
        return { _type: 'keyword', value: 'symbol' };
    }
    // Date
    if (isDate(type, libraryDescriptorName)) {
        logger.debug('Date');
        return { _type: 'class', value: ts__default.createIdentifier(libraryDescriptorName) };
    }
    // RegExp
    if (isRegExp(type, libraryDescriptorName)) {
        logger.debug('RegExp');
        return { _type: 'class', value: ts__default.createIdentifier(libraryDescriptorName) };
    }
    // Null
    if (isNull(type))
        return { _type: 'literal', value: ts__default.createNull() };
    // Undefined, Void
    if (isUndefined(type)) {
        logger.debug('Undefined');
        return { _type: 'literal', value: ts__default.createIdentifier('undefined') };
    }
    // Any
    if (isAny(type))
        return { _type: 'unspecified' };
    // Never
    if (isNever(type))
        return { _type: 'never' };
    // For the checks below we need access to the TypeNode for this type
    const typeNode = typeChecker.typeToTypeNode(type, scope);
    const typeName = typeChecker.typeToString(type, scope);
    // True
    if (isTrueKeyword(typeNode))
        return { _type: 'literal', value: ts__default.createTrue() };
    // False
    if (isFalseKeyword(typeNode))
        return { _type: 'literal', value: ts__default.createFalse() };
    // Promise
    //
    // Checking promises is not precise - it is not possible to type-check
    // the resolution value. A warning is printed notifying the consumer about this.
    if (isPromise(type, libraryDescriptorName)) {
        logger.debug('Promise');
        logger.warn(promiseTypeWarning(typeName));
        return (resolve) => ({
            _type: 'promise',
            properties: getPropertyTypeDescriptors(typeChecker, scope, type.getProperties(), resolve),
        });
    }
    // Literal types
    if (isLiteral(type)) {
        logger.debug('Literal');
        const value = type.value;
        if (value === undefined) {
            if (type.getFlags() & ts__default.TypeFlags.BooleanLiteral) {
                if (typeName === 'false')
                    return { _type: 'literal', value: ts__default.createLiteral(false) };
                if (typeName === 'true')
                    return { _type: 'literal', value: ts__default.createLiteral(true) };
            }
            throw new Error('Could not find value for a literal type ' + typeName);
        }
        return { _type: 'literal', value: ts__default.createLiteral(value) };
    }
    // Intersection
    if (isIntersection(type, typeNode)) {
        logger.debug('Intersection type');
        return (resolve) => ({
            _type: 'intersection',
            types: type.types.map((type) => resolve(scope, type)),
        });
    }
    // Union
    if (isUnion(type, typeNode)) {
        logger.debug('Union type');
        return (resolve) => ({
            _type: 'union',
            types: type.types.map((type) => resolve(scope, type)),
        });
    }
    // object
    if (isObjectKeyword(typeNode)) {
        logger.debug('object (keyword)');
        return { _type: 'keyword', value: 'object' };
    }
    // Tuple
    if (isTuple(type, typeNode)) {
        logger.debug('Tuple');
        const typeArguments = type.typeArguments || [];
        return (resolve) => ({
            _type: 'tuple',
            types: typeArguments.map((type) => resolve(scope, type)),
        });
    }
    // Function
    if (isFunction(type, libraryDescriptorName, typeNode)) {
        logger.debug('Function');
        logger.info(functionTypeWarning(typeName));
        const numberIndexType = type.getNumberIndexType();
        const stringIndexType = type.getStringIndexType();
        // If the type refers to Function library type we need to skip property checks
        // since properties of Function include e.g. apply, call or bind which are all of type Function.
        // Leaving them in would create an infinite loop.
        //
        // This though is not unsafe since checking whether typeof value === 'function'
        // is enough to be certain that these properties exist.
        //
        // If the type refers to e.g. a function literal we use type.getProperties()
        // which will only return the explicitly defined properties (skipping apply, call, bind etc)
        const properties = libraryDescriptorName === 'Function' ? [] : type.getProperties();
        return (resolve) => ({
            _type: 'function',
            properties: getPropertyTypeDescriptors(typeChecker, scope, properties, resolve),
            numberIndexType: numberIndexType ? resolve(scope, numberIndexType) : undefined,
            stringIndexType: stringIndexType ? resolve(scope, stringIndexType) : undefined,
        });
    }
    // Array
    if (isArray(typeChecker, type, libraryDescriptorName, typeNode)) {
        logger.debug('Array');
        const elementType = (_a = type.typeArguments) === null || _a === void 0 ? void 0 : _a[0];
        if (!elementType) {
            throw new Error('Could not find element type for (apparently) array type ' + typeName);
        }
        return (resolve) => ({ _type: 'array', type: resolve(scope, elementType) });
    }
    // Map
    if (isMap(type, libraryDescriptorName)) {
        logger.debug('Map');
        const [keyType, valueType] = type.typeArguments || [];
        if (!keyType) {
            throw new Error('Could not find key type for (apparently) Map type ' + typeName);
        }
        if (!valueType) {
            throw new Error('Could not find value type for (apparently) Map type ' + typeName);
        }
        return (resolve) => ({
            _type: 'map',
            keyType: resolve(scope, keyType),
            valueType: resolve(scope, valueType),
        });
    }
    // Set
    if (isSet(type, libraryDescriptorName)) {
        logger.debug('Set');
        const [setType] = type.typeArguments || [];
        if (!setType) {
            throw new Error('Could not find key type for (apparently) Set type ' + typeName);
        }
        return (resolve) => ({ _type: 'set', type: resolve(scope, setType) });
    }
    // DOM Element / Node
    const domElementClassName = getDOMElementClassName(program, type);
    if (domElementClassName) {
        logger.debug('DOM');
        return { _type: 'class', value: ts__default.createIdentifier(domElementClassName) };
    }
    // Interface-ish
    if (isInterface(type, libraryDescriptorName)) {
        logger.debug('Interface');
        const numberIndexType = type.getNumberIndexType();
        const stringIndexType = type.getStringIndexType();
        return (resolve) => ({
            _type: 'interface',
            properties: getPropertyTypeDescriptors(typeChecker, scope, type.getApparentProperties(), resolve),
            numberIndexType: numberIndexType ? resolve(scope, numberIndexType) : undefined,
            stringIndexType: stringIndexType ? resolve(scope, stringIndexType) : undefined,
        });
    }
    throw new Error('Unable to describe type ' + typeName);
};

const createArrayTypeGuard = (value, createElementCheck) => createLogicalAndChain(createIsArray(value), createArrayEvery(value, createElementCheck));
const createSetTypeGuard = (value, createElementCheck) => {
    const setValues = ts__default.createCall(ts__default.createPropertyAccess(value, 'values'), [], []);
    const setValuesAsArray = createArrayFrom(setValues);
    const elementChecks = createArrayEvery(setValuesAsArray, createElementCheck);
    return createLogicalAndChain(createIsInstanceOf(value, ts__default.createIdentifier('Set')), elementChecks);
};
const createMapTypeGuard = (value, createKeyCheck, createValueCheck) => {
    const mapEntries = ts__default.createCall(ts__default.createPropertyAccess(value, 'entries'), [], []);
    const mapEntriesAsArray = createArrayFrom(mapEntries);
    const entryChecks = createArrayEvery(mapEntriesAsArray, (entry) => createLogicalAndChain(createKeyCheck(ts__default.createElementAccess(entry, 0)), createValueCheck(ts__default.createElementAccess(entry, 1))), 'entry');
    return createLogicalAndChain(createIsInstanceOf(value, ts__default.createIdentifier('Map')), entryChecks);
};
const createTupleTypeGuard = (value, length, createElementCheck) => {
    const arrayLengthCheck = ts__default.createStrictEquality(ts__default.createPropertyAccess(value, 'length'), ts__default.createLiteral(length));
    const elementChecks = Array.from({ length }).map((_, index) => createElementCheck(ts__default.createElementAccess(value, index), index));
    return createLogicalAndChain(createIsArray(value), arrayLengthCheck, ...elementChecks);
};
const createIsNotNumeric = (value) => createLogicalAndChain(ts__default.createCall(ts__default.createIdentifier('isNaN'), [], [ts__default.createCall(ts__default.createIdentifier('parseFloat'), [], [value])]), ts__default.createStrictInequality(value, ts__default.createLiteral('NaN')));
const createObjectTypeGuard = (value, { properties, numberIndexType, stringIndexType }, typeGuardGenerator) => {
    const objectKeys = createObjectKeys(value);
    const basicTypeGuard = stringIndexType || numberIndexType ? createIsNotPrimitive(value) : createIsNotNullOrUndefined(value);
    const propertyChecks = properties.map(({ type, accessor }) => typeGuardGenerator(type, ts__default.createElementAccess(value, accessor)));
    const indexPropertyChecks = stringIndexType || numberIndexType
        ? [
            createArrayEvery(objectKeys, (key) => {
                const propertyValue = ts__default.createElementAccess(value, key);
                const numberPropertyCheck = numberIndexType
                    ? createLogicalOrChain(createIsNotNumeric(key), typeGuardGenerator(numberIndexType, propertyValue))
                    : undefined;
                const stringPropertyCheck = stringIndexType
                    ? typeGuardGenerator(stringIndexType, propertyValue)
                    : undefined;
                const checks = [numberPropertyCheck, stringPropertyCheck].filter(Boolean);
                return createLogicalAndChain(...checks);
            }, 'key'),
        ]
        : [];
    return createLogicalAndChain(basicTypeGuard, ...propertyChecks, ...indexPropertyChecks);
};

const createTypeGuardGenerator = (typeGuardRegistry, typeDescriptorRegistry, strictNullChecks) => {
    // If the strict null checks are off (which they shouldn't),
    // the optional types all pretend to be non-optional
    // so we need to make sure every type can also be null or undefined
    const prefixNullChecks = strictNullChecks
        ? // If strictNullChecks are on just return the original type guard
            (typeGuard) => typeGuard
        : // But if they are off prepend every type guard with check for null or undefined
            (typeGuard, value) => {
                return createLogicalOrChain(ts__default.createStrictEquality(value, ts__default.createNull()), ts__default.createStrictEquality(value, ts__default.createIdentifier('undefined')), typeGuard);
            };
    const typeGuardGenerator = (typeName, value) => {
        // Step 1: Check if type guard already exists for this type
        const storedTypeGuardFunction = typeGuardRegistry.get(typeName);
        if (storedTypeGuardFunction) {
            return ts__default.createCall(storedTypeGuardFunction, [], [value]);
        }
        // Step 2: Get the TypeDescriptor for the typeName
        const typeDescriptor = typeDescriptorRegistry.get(typeName);
        if (typeDescriptor === undefined) {
            throw new Error(`Unable to find type descriptor for type '${typeName}'`);
        }
        // Step 3: Turn the TypeDescriptor into a type guard expression
        //
        // In this step we need to prefix all applicable type guards
        // with null checks bypass (see above)
        switch (typeDescriptor._type) {
            case 'literal':
                return prefixNullChecks(ts__default.createStrictEquality(value, typeDescriptor.value), value);
            case 'keyword':
                switch (typeDescriptor.value) {
                    case 'object':
                        return prefixNullChecks(createIsNotPrimitive(value), value);
                    default:
                        return prefixNullChecks(ts__default.createStrictEquality(ts__default.createTypeOf(value), ts__default.createLiteral(typeDescriptor.value)), value);
                }
            case 'intersection':
                const intersectionTypeCheckMethod = typeGuardRegistry.create(typeName, (value) => prefixNullChecks(createLogicalAndChain(...typeDescriptor.types.map((typeName) => typeGuardGenerator(typeName, value))), value));
                return ts__default.createCall(intersectionTypeCheckMethod, undefined, [value]);
            case 'union':
                const unionTypeCheckMethod = typeGuardRegistry.create(typeName, (value) => prefixNullChecks(createLogicalOrChain(...typeDescriptor.types.map((typeName) => typeGuardGenerator(typeName, value))), value));
                return ts__default.createCall(unionTypeCheckMethod, undefined, [value]);
            case 'array':
                return prefixNullChecks(createArrayTypeGuard(value, (element) => typeGuardGenerator(typeDescriptor.type, element)), value);
            case 'tuple':
                return prefixNullChecks(createTupleTypeGuard(value, typeDescriptor.types.length, (element, index) => {
                    return typeGuardGenerator(typeDescriptor.types[index], element);
                }), value);
            case 'class':
                return prefixNullChecks(createIsInstanceOf(value, typeDescriptor.value), value);
            case 'map':
                return prefixNullChecks(createMapTypeGuard(value, (key) => typeGuardGenerator(typeDescriptor.keyType, key), (value) => typeGuardGenerator(typeDescriptor.valueType, value)), value);
            case 'set':
                return prefixNullChecks(createSetTypeGuard(value, (element) => typeGuardGenerator(typeDescriptor.type, element)), value);
            case 'promise':
                const promiseTypeCheckMethod = typeGuardRegistry.create('Promise', (value) => prefixNullChecks(createObjectTypeGuard(value, { properties: typeDescriptor.properties }, typeGuardGenerator), value));
                return ts__default.createCall(promiseTypeCheckMethod, undefined, [value]);
            case 'function':
                const functionTypeCheck = ts__default.createStrictEquality(ts__default.createTypeOf(value), ts__default.createLiteral('function'));
                // If this is just a simple function with no additional properties then return the typeof check immediately
                if (!typeDescriptor.properties.length && !typeDescriptor.stringIndexType && !typeDescriptor.numberIndexType) {
                    return functionTypeCheck;
                }
                // If though the function has additional properties and/or index we need to check that
                const functionTypeCheckMethod = typeGuardRegistry.create(typeName, (value) => prefixNullChecks(createLogicalAndChain(functionTypeCheck, createObjectTypeGuard(value, typeDescriptor, typeGuardGenerator)), value));
                return ts__default.createCall(functionTypeCheckMethod, undefined, [value]);
            case 'interface':
                const objectTypeCheckMethod = typeGuardRegistry.create(typeName, (value) => prefixNullChecks(createObjectTypeGuard(value, typeDescriptor, typeGuardGenerator), value));
                return ts__default.createCall(objectTypeCheckMethod, undefined, [value]);
            case 'unspecified':
                return ts__default.createTrue();
            case 'never':
                return ts__default.createFalse();
            default:
                throw new Error('Unable to create a checker for type descriptor ' + typeDescriptor._type);
        }
    };
    return typeGuardGenerator;
};

const createTypeGuardResolver = (program, typeNameResolver, typeGuardGenerator) => {
    const typeChecker = program.getTypeChecker();
    const typeGuardResolver = (typeNode, value) => {
        // Step 1: Get type from TypeNode
        const type = typeChecker.getTypeFromTypeNode(typeNode);
        // Step 2: Resolve the type descriptor
        const typeName = typeNameResolver(typeNode, type);
        // Step 3: Turn the type descriptor into a type guard
        const typeGuard = typeGuardGenerator(typeName, value);
        // Step 4: RETURN!!!
        return typeGuard;
    };
    return typeGuardResolver;
};

const createTypeNameResolver = (registry, typeDescriptorGenerator) => {
    const typeNameResolver = (scope, type) => {
        // FIXME Maybe the registry does not need to pass typeName as the first parameter to create factory
        const typeName = registry.create(type, () => {
            const typeDescriptorOrCallback = typeDescriptorGenerator(scope, type);
            const typeDescriptor = typeof typeDescriptorOrCallback === 'function'
                ? typeDescriptorOrCallback(typeNameResolver)
                : typeDescriptorOrCallback;
            return typeDescriptor;
        });
        return typeName;
    };
    return typeNameResolver;
};

function visitNode(node, visitor) {
    if (ts__default.isSourceFile(node))
        return node;
    return visitor(node);
}
function transformUsingVisitor(node, context, visitor) {
    return ts.visitEachChild(visitNode(node, visitor), (childNode) => transformUsingVisitor(childNode, context, visitor), context);
}

/**
 * The main transformer function.
 *
 * This needs to be registered as a TypeScript "before" transform
 * in your build/test configuration.
 *
 * See https://www.npmjs.com/package/ts-type-checked#installation for more information
 *
 * @param program {ts.Program} An instance of TypeScript Program
 * @param options {Partial<TransformerOptions>} Transformer options object
 */
var index = (program, options = {}) => {
    const resolvedOptions = Object.assign(Object.assign({}, defaultTransformerOptions), options);
    const { mode, logLevel } = resolvedOptions;
    // Without strict null checks on we need to
    const compilerOptions = program.getCompilerOptions();
    const strictNullChecks = compilerOptions.strictNullChecks === undefined ? !!compilerOptions.strict : !!compilerOptions.strictNullChecks;
    // Get a reference to a TypeScript TypeChecker in order to resolve types from type nodes
    const typeChecker = program.getTypeChecker();
    return (context) => (file) => {
        // Create a file specific logger
        const logger = createLogger(logLevel, `[${file.fileName}]`);
        // We will need a utility that can generate unique type names
        // that can be used to reference type descriptors and type guards.
        //
        // Since we are doing the transformation on per-file basis we will create
        // it here so that the type names are unique within one file but not necessarily
        // unique across the codebase
        const typeNameGenerator = mode === 'development' ? createDebugTypeNameGenerator(typeChecker) : createProductionTypeNameGenerator();
        // Now we need to walk the AST and replace all the references to isA or typeCheckFor
        // with generated type guards.
        //
        // Simple type guards (string, number etc.) will be inlined since they translate
        // into something like
        //
        // value => typeof value === 'string'
        //
        // But the interface type guards could contain cycles (on the type level, e.g.)
        //
        // interface Node {
        //   parent?: Node;
        // }
        //
        // If we inlined these the code generator would be spinning in cycles. Therefore
        // we need to store them somewhere to be able to access them by the unique type id
        // so that they can call themselves recursively.
        //
        // The solution is to create a runtime variable, a hash with type names
        // as keys and type guards as values, let's call it type guard map:
        const typeGuardRegistryIdentifier = ts__default.createIdentifier('___isA___');
        const typeGuardRegistry = new HashTypeGuardRegistry(typeGuardRegistryIdentifier);
        const typeDescriptorRegistry = new MapTypeDescriptorRegistry(typeNameGenerator);
        const typeGuardGenerator = createTypeGuardGenerator(typeGuardRegistry, typeDescriptorRegistry, strictNullChecks);
        const typeDescriptorCreator = createTypeDescriptorGenerator(program, logger);
        const typeNameResolver = createTypeNameResolver(typeDescriptorRegistry, typeDescriptorCreator);
        const typeGuardResolver = createTypeGuardResolver(program, typeNameResolver, typeGuardGenerator);
        // Now we need an ASTVisitor function that will replace all occurrences
        // of isA and typeCheckFor with the generated type guards
        const visitor = createTypeCheckVisitor(typeChecker, typeGuardResolver);
        // We use the visitor to transform the file
        const transformedFile = transformUsingVisitor(file, context, visitor);
        // The last step is to insert the type guard hash into the source file
        return ts__default.updateSourceFileNode(transformedFile, [...typeGuardRegistry.code(), ...transformedFile.statements]);
    };
};

module.exports = index;
